   
# gtUtils.py
# Andrew Davison, ad@coe.psu.ac.th, July 2024

'''
Game Theory 101 playlist
https://www.youtube.com/watch?v=NSVmOC_5zrE&list=PLKI1h_nAkaQoDzI4xDIXzx6U2ergFmedo

Simultaneous games 1-15, 27-37
'''

import matplotlib
import matplotlib.pyplot as plt
import itertools
from copy import deepcopy
from player import Player



def readPayoff(fnm):
  '''
    Ignore comment lines and blank lines.
    The first line is the size of the square payoff matrix.
    The second line is the game title.
    Subsequent lines are rows of the payoff matrix, of the form:
       label  p1-score p2-score  ...  (size pairs of scores)

  '''
  clabels = [None]  # will hold the title, and row/col labels
  payoffs = [None]
  lineCount = 0
  payRow = 0   # the current row index of the payoff matrix
  print("Reading", fnm)
  for line in open(fnm):
    li = line.strip()
    if li != "" and not li.startswith("#"):
      if lineCount == 0:
        size = int(li)   # size of the square payoff matrix
        print("Building", size, "x", size, "payoff...")
        clabels = [""]*(size+1)
        payoffs = [None]*size
        for r in range(size):
          payoffs[r] = [None]*size
          for c in range(size):
             payoffs[r][c] = (0,0)
        lineCount += 1
      elif lineCount == 1:
        clabels[0] = li   # game title
        lineCount += 1
      elif lineCount == 2:
        args = li.split()
        clabels[payRow+1] = args[0]
        pos = 1
        for c in range(size):    # read in a row of the payoff matrix
          payoffs[payRow][c] = ( int(args[pos]), int(args[pos+1]))  # (P1, P2)
          pos += 2
        payRow += 1
  print("Labels:", clabels)
  print("Payoffs:", payoffs)
  return clabels, payoffs   


def isSquare(arr):
  return all([len(row)==len(arr) for row in arr])

# --------------------------------------------

def playGames(payoffs, numPlayers, numTrials):
  stratsP1, stratsP2 = \
    buildStrategies(payoffs, numPlayers, numTrials)
  # print("stratsP1", stratsP1)
  return collectPoints(stratsP1, stratsP2, 
                          numPlayers, len(payoffs))



def buildStrategies(payoffs, numPlayers, numTrials):
  '''
    Tally the mix of play strategies for all the players over
    numTrials games. Each player plays everyone else, 
    except themselves, in a trial.

    The tally of player strategies at the end of one trial is
    recorded by incrementing an n-dimensional array cell that
    is indexed by those counts. 

    This will build up into a record of the most popular mix of
    play strategies by all the players.
  '''
  nPrefs = len(payoffs)
  sz = [numPlayers+1]*nPrefs
  stratsP1 = ndarray(sz, val=0)
  stratsP2 = ndarray(sz, val=0)
  players = [ Player(i, nPrefs) for i in range(numPlayers)]
  for t in range(numTrials):
    playAll(players, payoffs)
    sIdxs = updateStrats(players, 1, nPrefs)  # when P1
      # e.g. returns [2,3,1,4] mix of 10 players for 4 prefs
    incrCell(stratsP1, sIdxs)
      # e.g. add 1 to stratsP1[2][3][1][4]
    sIdxs = updateStrats(players, 2, nPrefs)  # when P2
    incrCell(stratsP2, sIdxs)
  return stratsP1, stratsP2



def ndarray(dims, val=None):
  # dims = [d0, d1, ..., dn-1]
  # create n-dim array/list of size d0 x d1 x ... dn-1
  # https://stackoverflow.com/questions/71041205/is-it-possible-to-define-function-that-creates-an-n-dimensional-list-of-lists-ea
  base = val
  for d in reversed(dims):
    base = list(map(deepcopy, itertools.repeat(base, times=d)))
  return base


def playAll(players, payoffs):
  # record the scores for all the players vs. each other
  # using their preferred strategy
  n = len(players)
  for i in range(n):
    p1St = players[i].getStrategy(1)
    for j in range(n):
      if players[i].getId != players[j].getId():
        # print(i, j)
        p2St = players[j].getStrategy(2)
        p1Pay, p2Pay = payoffs[p1St][p2St]
        players[i].addScore(p1Pay, 1)
        players[j].addScore(p2Pay, 2)


def updateStrats(players, order, nPrefs):
  '''
    tally the current play strategies for all the players
    when playing first or second, and calculate the average
    score for all the players, so their preference weights
    can be adjusted
  '''
  currStrategies = [0]*nPrefs    
       # tally of play strategies used by all players
  totScore = 0
  for p in players: 
    currStrategies[p.getStrategy(order)] += 1
    totScore += p.getScore(order)
  avgScore = totScore/len(players)
  for p in players:
    # adjust the preference weightings based on how
    # the recorded score exceeds/falls short of the average 
    p.updatePrefs(avgScore, order)
  return currStrategies


def incrCell(arr, idxs):
  # increment value in arr[][]...
  idx = idxs[0]
  if len(idxs) == 1:
    arr[idx] += 1
  else:
    incrCell(arr[idx], idxs[1:])


# ----------------------------------------------


def collectPoints(stratsP1, stratsP2, 
                     numPlayers, nPrefs, dotSize=2):
  ''' 
    collect points for P1 and P2
    Each point represents a single mix of player strategies.
    The popularity of that mix is reflected in the size of the point.
  '''
  pts1 = []; pts2 = []
  szs1 = []; szs2 = []
  for idxs in ndRange(numPlayers+1, nPrefs):
    # iterate through every cell of the strats arrays
    sz = getCell(stratsP1, idxs)
      # e.g. get size in stratsP1[2][3][1][4]
      # e.g. represents [2,3,1,4] mix of 10 players for 4 prefs
      # and size is how many times this mix occurred
    probs = [ idx/numPlayers for idx in idxs]
    if sz != 0:
      pts1.append(probs)
      szs1.append(sz*dotSize)
    sz = getCell(stratsP2, idxs)
    if sz != 0:
      pts2.append(probs)
      szs2.append(sz*dotSize)
  return pts1, szs1, pts2, szs2


def ndRange(size, n):
  # generate all indicies for n-dim array of size^n
  if not n:
    yield ()
    return
  for outer in ndRange(size, n-1):
    for inner in range(size):
      yield outer + (inner,)


def getCell(arr, idxs):
  # idxs = [idx0, idx1, ...]
  # get value in arr[idx0][idx1]...
  idx = idxs[0]
  if len(idxs) == 1:
    return arr[idx]
  else:
    return getCell(arr[idx], idxs[1:])

# ----------------------------------------------

def baryToCart(barys, carts, scale=1):    
  # transform from barycentric space to cartesian coords
  # https://en.wikipedia.org/wiki/Barycentric_coordinate_system
  lls = transpose(carts)
  return dotArr(lls, barys, scale)
  return dArr


def transpose(arr):
  return [[row[i] for row in arr] for i in range(len(arr[0]))]


def dotArr(arr1, arr2, scale):
  nRows1 = len(arr1)
  nRows2 = len(arr2)
  dArr = [[0 for i in range(nRows1)] for j in range(nRows2)]
  for r2 in range(nRows2):
    for r1 in range(nRows1):
      dArr[r2][r1] = ldot(arr1[r1], arr2[r2], scale)
  return dArr


def ldot(l1, l2, scale):
  return sum([i*j/scale for (i,j) in zip(l1, l2)])


# ----------------------------------------------

def wCoords(p, carts):
  # transform a cartesian coord to barycentric space
  sz = len(carts)
  wSum = sum([wachspress(i,p,carts) for i in range(sz)])
  return [ wachspress(i,p,carts)/wSum for i in range(sz)]


def wachspress(i, p, carts):
  sz = len(carts)
  a = carts[(i-1)%sz]
  b = carts[i]
  c = carts[(i+1)%sz]
  return triArea(a,b,c)/(triArea(a,b,p) * triArea(b,c,p))


def triArea(a, b, c):
  return 0.5 * (a[0]*(b[1]-c[1]) + 
               b[0]*(c[1]-a[1]) +
               c[0]*(a[1]-b[1]))


# ----------------------------------------------

def setWindow(plt, title, x, y):
  # position named matplotlib window on the desktop
  f = plt.gcf()
  f.canvas.manager.set_window_title(title)
  backend = matplotlib.get_backend()
  if backend == 'TkAgg':
    f.canvas.manager.window.wm_geometry("+%d+%d" % (x, y))
  elif backend == 'WXAgg':
    f.canvas.manager.window.SetPosition((x, y))
  else:  # This works for QT and GTK
    f.canvas.manager.window.move(x, y)


