   
# player.py
# Andrew Davison, ad@coe.psu.ac.th, July 2024
'''
  Game theory player.
  A player (with id) can be both first player (P1) and the
  second player (P2).

  On each turn, a play strategy is chosen based on weighted
  preferences which are adjusted based on the payoff scores
  from previous plays
'''

import random, math


PREF_WEIGHT = 100
MIN_WEIGHT = PREF_WEIGHT/2


class Player:

  def __init__(self, id, nPrefs):
    self.id = id
    # when player is P1
    self.strategy1 = None
    self.prefs1 = [PREF_WEIGHT]*nPrefs
       # weights used to choose a play preference
    self.score1 = 0  # payoff from the chosen play
    # print(self.prefs1)

    # when player is P2
    self.strategy2 = None
    self.prefs2 = [PREF_WEIGHT]*nPrefs
    self.score2 = 0


  def getId(self):
    return self.id

  def getStrategy(self, order):
    # randomly choose a play strategy using the preference weights
    if order == 1:   # P1
      if self.strategy1 == None:
        self.strategy1 = random.choices(
                            range(len(self.prefs1)), 
                            weights=self.prefs1)[0]
      return self.strategy1
    else:  # P2
      if self.strategy2 == None:
        self.strategy2 = random.choices(
                            range(len(self.prefs2)), 
                            weights=self.prefs2)[0]
      return self.strategy2


  def addScore(self, s, order):
    # a score may be positive or negative
    if order == 1:  # P1
      self.score1 += s
    else:  # P2
      self.score2 += s


  def getScore(self, order):
    return self.score1 if order == 1 else self.score2


  def updatePrefs(self, avgScore, order):
    '''
      adjust preference weight for the current play
      strategy based on difference of 
      score from the average score for all the players. 
      There is a minimum weight.
    '''
    if order == 1:  # P1
      s = math.floor(self.score1 - avgScore) 
        # score may be postive or negative
      self.prefs1[self.strategy1] += s
      if self.prefs1[self.strategy1] < MIN_WEIGHT:
        self.prefs1[self.strategy1] = MIN_WEIGHT
      # discard current score and play strategy
      # to reset the player for the next trial
      self.score1 = 0
      self.strategy1 = None
    else:   # P2
      s = math.floor(self.score2 - avgScore)
      self.prefs2[self.strategy2] += s
      if self.prefs2[self.strategy2] < MIN_WEIGHT:
        self.prefs2[self.strategy2] = MIN_WEIGHT
      # reset
      self.score2 = 0
      self.strategy2 = None


  def __str__(self):
   return f"<{self.id}: {self.prefs1} {self.prefs2}>"
