
# simNN.py
# Andrew Davison, ad@coe.psu.ac.th, July 2024

'''
  Simulate games with a nxn payoff matrix (n > 2).

  Display points inside a regular polygon using Barycentric coordinates.

  Example files:
    - see 
        sim33.py (fictious.txt, rps3.txt, rps36.txt, rps37.txt)
        sim44.py  (duoComp.txt, generals.txt, randomGame.txt, rowCol.txt)
    - duoComp6.txt

# https://coderslegacy.com/python/drawing-shapes-matplotlib-patches/
# https://matplotlib.org/stable/api/_as_gen/matplotlib.patches.RegularPolygon.html
'''


import math, sys
import matplotlib.pyplot as plt
from matplotlib.patches import RegularPolygon
import gtUtils


NUM_SESSIONS = 10
NUM_ROWS = 2
NUM_TRIALS = 500
NUM_PLAYERS = 10


def onPick(event):
  x = event.mouseevent.xdata
  y = event.mouseevent.ydata
  # print("Axes x, y:", x, y)
  # print("label", event.artist.get_label())
  bcrds = gtUtils.wCoords([x,y], verts)
  print("bary coords:", [float(f"{c:.2f}") for c in bcrds])
  # print("ind", event.ind[0])


def toLabel(lbls):
  s = "["
  for i in range(len(lbls)-1):
    s += str(lbls[i]) + ","
  s += str(lbls[-1]) + "]"
  return s


def baryChart(plt):
  plt.xlim(-1, 1)
  plt.ylim(-1, 1)
  plt.axis('scaled')
  plt.axis('off')
  
  for i in range(nPrefs):
    plt.plot([0, verts[i][0]], [0, verts[i][1]], 'r', ls=":")
    plt.text(verts[i][0], verts[i][1], toLabel(labels[i]), 
                          fontsize="small", ha="center")
  
  shape = RegularPolygon((0, 0), numVertices=nPrefs,
             radius=1, orientation=angle-math.pi/2,
             edgecolor="blue")
  shape.set_fill(False)
  
  ax = plt.gca()
  ax.add_patch(shape)
  plt.text(0.01, 0.99, ','.join(clabels[1:]), 
          ha='left', va='top', transform=ax.transAxes)
  plt.title(clabels[0], pad=20)


def plotPlayers(pts1, szs1, pts2, szs2):
  # scatter plots
  # P1
  cpts1 = gtUtils.baryToCart(pts1, verts)
  xs1, ys1 = zip(*cpts1)
  plt.scatter(xs1, ys1, s=szs1, alpha=0.5, label="P1", picker=10)  
  # P2
  cpts2 = gtUtils.baryToCart(pts2, verts)
  xs2, ys2 = zip(*cpts2)
  plt.scatter(xs2, ys2, s=szs2, alpha=0.5, label="P2", picker=10) 
  plt.legend()



# ----------------------

if len(sys.argv) != 2:
  print("Usage: python simNN.py <fnm>")
  sys.exit()

fnm = sys.argv[1]
clabels, payoffs = gtUtils.readPayoff(fnm)

if not gtUtils.isSquare(payoffs):
  print("Payoff matrix must be square")
  sys.exit()

nPrefs = len(payoffs)
if nPrefs <= 2:
  print("Too few preferences (try sim22.py)")
  sys.exit()

angle = 2*math.pi/nPrefs
verts = [ (math.cos(i*angle), math.sin(i*angle)) for i in range(nPrefs)]
labels = [None]*nPrefs
for i in range(nPrefs):
  labels[i] = [0]*nPrefs
  labels[i][i] = 1

fig = plt.figure(figsize=(3.4*NUM_SESSIONS//NUM_ROWS, 4*NUM_ROWS))
plt.tight_layout()
plt.subplots_adjust(left=0.05)
gtUtils.setWindow(plt, fnm + ": " + clabels[0], 100,100)

for i in range(NUM_SESSIONS):
  plt.subplot(NUM_ROWS, NUM_SESSIONS//NUM_ROWS, i+1)
  baryChart(plt)
  pts1, szs1, pts2, szs2 = \
    gtUtils.playGames(payoffs, NUM_PLAYERS, NUM_TRIALS)
  plotPlayers(pts1, szs1, pts2, szs2)

  fig = plt.gcf()
  fig.canvas.callbacks.connect('pick_event', onPick)

plt.show()
